#!/usr/bin/env python

import logging
import logging.handlers
from optparse import OptionParser
import os
import sys
import urlparse

from configobj import ConfigObj

from paste import httpserver
from paste.httpexceptions import make_middleware
from paste.translogger import make_filter

from loggerhead import daemon, release
from loggerhead.apps.config import Root


def make_handler(config, filename):
    roll = config.get('log.roll', 'never')
    if roll == 'daily':
        h = logging.handlers.TimedRotatingFileHandler(filename, 'midnight', 0, 100)
    elif roll == 'weekly':
        h = logging.handlers.TimedRotatingFileHandler(filename, 'W0', 0, 100)
    else:
        h = logging.FileHandler(filename)
    return h


def setup_logging(home, config, foreground):
    # i hate that stupid logging config format, so just set up logging here.

    log_folder = os.path.join(home, 'logs')
    if not os.path.exists(log_folder):
        os.mkdir(log_folder)

    f = logging.Formatter('%(levelname)-.3s [%(asctime)s.%(msecs)03d] %(name)s: %(message)s',
                          '%Y%m%d-%H:%M:%S')
    debug_log = make_handler(config, os.path.join(log_folder, 'debug.log'))
    debug_log.setLevel(logging.DEBUG)
    debug_log.setFormatter(f)
    if foreground:
        stdout_log = logging.StreamHandler(sys.stdout)
        stdout_log.setLevel(logging.DEBUG)
        stdout_log.setFormatter(f)
    f = logging.Formatter('[%(asctime)s.%(msecs)03d] %(message)s',
                          '%Y%m%d-%H:%M:%S')
    access_log = make_handler(config, os.path.join(log_folder, 'access.log'))
    access_log.setLevel(logging.INFO)
    access_log.setFormatter(f)

    logging.getLogger('').addHandler(debug_log)
    logging.getLogger('turbogears.access').addHandler(access_log)
    logging.getLogger('turbogears.controllers').setLevel(logging.INFO)

    if foreground:
        logging.getLogger('').addHandler(stdout_log)



def main():
    parser = OptionParser(usage='usage: %prog [options]', version='%prog ' + release.version)
    parser.add_option('-f', '--foreground', action='store_true', dest='foreground', default=False,
                      help="run in the foreground; don't daemonize")
    parser.add_option('-C', '--check', action='store_true', dest='check', default=False,
                      help="only start if not already running (useful for cron jobs)")
    options, args = parser.parse_args()
    if len(args) > 0:
        parser.error('No filename arguments are used, only options.')


    home = os.path.realpath(os.path.dirname(__file__))
    pidfile = os.path.join(home, 'loggerhead.pid')

    if options.check:
        if daemon.is_running(pidfile):
            sys.exit(0)
        sys.stderr.write('Did not find loggerhead running in %r; restarting...\n' % (pidfile,))

    # read loggerhead config

    config = ConfigObj(os.path.join(home, 'loggerhead.conf'), encoding='utf-8')
    extra_path = config.get('bzrpath', None)
    if extra_path:
        sys.path.insert(0, extra_path)

    #turbogears.update_config(configfile="dev.cfg", modulename="loggerhead.config")

    potential_overrides = [ ('server.socket_port', int),
                            ('server.webpath', str),
                            ('server.thread_pool', int),
                            ('server.socket_host' ,str) ]
    server_port = int(config.get('server.socket_port', 8080))
    nworkers = int(config.get('server.thread_pool', 10))
    server_host = config.get('server.socket_host', '0.0.0.0')
    webpath = config.get('server.webpath', None)

    for key, keytype in potential_overrides:
        value = config.get(key, None)
        if value is not None:
            value = keytype(value)
            #turbogears.config.update({ key: value })

    if not options.foreground:
        sys.stderr.write('\n')
        sys.stderr.write('Launching loggerhead into the background.\n')
        sys.stderr.write('PID file: %s\n' % (pidfile,))
        sys.stderr.write('\n')

        daemon.daemonize(pidfile, home)

    setup_logging(home, config, foreground=options.foreground)

    log = logging.getLogger('loggerhead')
    log.info('Starting up...')

    app = Root(config)

    app = app
    app = make_middleware(app)
    app = make_filter(app, None)

    if webpath:
        scheme, netloc, path, blah, blah, blah = urlparse.urlparse(webpath)
        def app(environ, start_response, orig=app):
            environ['SCRIPT_NAME'] = path
            environ['HTTP_HOST'] = netloc
            return orig(environ, start_response)

    try:
        httpserver.serve(
            app, host=server_host, port=server_port,
            threadpool_workers=nworkers)
    finally:
        log.info('Shutdown.')
        try:
            os.remove(pidfile)
        except OSError:
            pass


if __name__ == '__main__':
    main()
